<div{{ field.attrs }}>
    <div class="ui toggle checkbox">
    {{ field.input }}
    {{ field.label }}
    {% if field.help_text %}
        <div class="ui pointing label">{{ field.help_text }}</div>
    {% endif %}
    {{ field.errors }}
    </div>
</div>
