<div{{ field.attrs }}>
    {{ field.label }}
    {{ field.input }}
    {% if field.value %}<img src="{{ field.value.url }}">{% endif %}
    <div class="helptext">{{ field.help_text }}</div>
    {{ field.errors }}
</div>
